#include "nrf_delay.h"
#include "LTC6811.h"
#include "dtiBMS.h"
#include "spi_global.h"

/**********************************************************
  Setup Variables
  The following variables can be modified to
  configure the software.
***********************************************************/
#define TOTAL_IC 1//!< Number of ICs in the daisy chain
#define TOTAL_CELL     12
#define COMPERE_VOLTAGE 10 //! 10 = 10mV

// 6811 Cell Fet switch
#define S1       0
#define S2       1
#define S3       2
#define S4       3
#define S5       4
#define S6       5
#define S7       6
#define S8       7
#define S9       8
#define S10      9
#define S11      10
#define S12      11

// 6811 Cell ADC Pin
#define C0       0
#define C1       1
#define C2       2
#define C3       3
#define C4       4
#define C5       5
#define C6       6
#define C7       7
#define C8       8
#define C9       9
#define C10      10
#define C11      11
#define C12      11

//Under Voltage and Over Voltage Thresholds
#define OV_THRESHOLD 41000 //!< Over voltage threshold ADC Code. LSB = 0.0001 ---(4.1V)
#define UV_THRESHOLD 30000 //!< Under voltage threshold ADC Code. LSB = 0.0001 ---(3V)

//Loop Measurement Setup. These Variables are ENABLED or DISABLED. Remember ALL CAPS
#define ENABLED 1
#define DISABLED 0

#define DATALOG_ENABLED 1
#define DATALOG_DISABLED 0

#define WRITE_CONFIG    DISABLED //!< Loop Measurement Setup
#define READ_CONFIG     DISABLED //!< Loop Measurement Setup
#define MEASURE_CELL    ENABLED //!< Loop Measurement Setup
#define MEASURE_AUX     DISABLED //!< Loop Measurement Setup
#define MEASURE_STAT    DISABLED //!< Loop Measurement Setup
#define PRINT_PEC       DISABLED //!< Loop Measurement Setup
/************************************
  END SETUP
*************************************/


/*********************************************************
 Set variables 
**********************************************************/
uint8_t bBalanceStatus = false;
uint16_t iCount;
uint8_t bStatusFET[11] = {false};

/*********************************************************
 Set the configuration bits. 
 Refer to the Configuration Register Group from data sheet. 
**********************************************************/
uint8_t REFON = true; //!< Reference Powered Up Bit
uint8_t ADCOPT = false; //!< ADC Mode option bit
uint8_t gpioBits_a[5] = {false,false,false,false,false}; //!< GPIO Pin Control // Gpio 1,2,3,4,5
uint16_t UV = UV_THRESHOLD; //!< Under-voltage Comparison Voltage
uint16_t OV = OV_THRESHOLD; //!< Over-voltage Comparison Voltage
uint8_t dccBits_a[12] = {false,false,false,false,false,false,false,false,false,false,false,false}; //!< Discharge cell switch //Dcc 1,2,3,4,5,6,7,8,9,10,11,12
uint8_t dctoBits[4] = {true, false, true, false}; //!< Discharge time value // Dcto 0,1,2,3 // Programed for 4 min 
/*Ensure that Dcto bits are set according to the required discharge time. Refer to the data sheet */


/******************************************************
 Global Battery Variables received from 681x commands.
 These variables store the results from the LTC6811
 register reads and the array lengths must be based
 on the number of ICs on the stack
 ******************************************************/
cell_asic bms_ic[TOTAL_IC]; //!< Global Battery Variable




void ltc6811_Init()
{
     LTC6811_init_cfg(TOTAL_IC, bms_ic);
    for (uint8_t current_ic = 0; current_ic<TOTAL_IC;current_ic++) 
    {
      LTC6811_set_cfgr(current_ic,bms_ic,REFON,ADCOPT,gpioBits_a,dccBits_a, dctoBits, UV, OV);
    }
    LTC6811_reset_crc_count(TOTAL_IC,bms_ic);
    LTC6811_init_reg_limits(TOTAL_IC,bms_ic);
    
    
    dischargerClear(); 
    
//            // Cell data
//    wakeup_sleep(TOTAL_IC);
//    LTC6811_wrcfg(TOTAL_IC,bms_ic);
}


void dischargerSet(uint8_t readIC)
{
    SEGGER_RTT_printf(0, "Cell[%d] discharge. \n", readIC);
    wakeup_sleep(TOTAL_IC);
    LTC6811_set_discharge(readIC, TOTAL_IC, bms_ic);
    LTC6811_wrcfg(TOTAL_IC, bms_ic);
    wakeup_idle(TOTAL_IC);
//    error = LTC6811_rdcfg(TOTAL_IC, bms_ic);
     LTC6811_rdcfg(TOTAL_IC, bms_ic);
//    check_error(error);
}

void dischargerClear()
{
    SEGGER_RTT_printf(0," #### FET(TR) Clear. #### \n");
    wakeup_sleep(TOTAL_IC);
    LTC6811_clear_discharge(TOTAL_IC, bms_ic);
    LTC6811_wrcfg(TOTAL_IC, bms_ic);
    wakeup_idle(TOTAL_IC);
//    error = LTC6811_rdcfg(TOTAL_IC, bms_ic);
    LTC6811_rdcfg(TOTAL_IC, bms_ic);
//    check_error(error);
}


void checkCell()
{
  SEGGER_RTT_printf(0, " #### BMS Check Cell #### \n");

  uint16_t iCellMax = bms_ic[0].cells.c_codes[0];
  uint16_t iCellMin = bms_ic[0].cells.c_codes[0];
  uint16_t iCellResult = 0;
  
  uint8_t iMin = 0;
  uint8_t iMax = 0;
  

  uint16_t iCompareVolt = 0;
  
  // Check min, max
  for(int i = 1; i < TOTAL_CELL - 1; i++)
  {
      if(i == C5) continue;
      if(iCellMax < bms_ic[0].cells.c_codes[i])         
      {
        iCellMax = bms_ic[0].cells.c_codes[i]; 
        iMax = i;
      }
      else if(iCellMin >= bms_ic[0].cells.c_codes[i])   
      {
        iCellMin = bms_ic[0].cells.c_codes[i];     
        iMin = 0;
      }
  }

  iCellMax = iCellMax / 10;
  iCellMin = iCellMin / 10;
  iCellResult =  iCellMax - iCellMin;
  
  SEGGER_RTT_printf(0, "Cell[MAX : %d - MIN : %d] = %d, Cell[Max] = %d, Cell[Min] = %d \n",iMax, iMin, iCellResult, iCellMax ,iCellMin);


  // Cell check
  if(iCellResult > COMPERE_VOLTAGE)   // COMPERE_VOLTAGE = 10mV Ҽ ڸ
  {
     for(int i = 0; i < TOTAL_CELL - 1; i++)
     {
        // LTC6811 C0-C12 without C5
        if(i == C5) continue; 
        
        if(iCellMin  + iCompareVolt <  bms_ic[0].cells.c_codes[i] / 10) 
          bStatusFET[i] = true;
        else
        {
          bStatusFET[i] = false;
        }                                   
     }
  }
  else
  {
    bBalanceStatus = false; // balancing off
    
    SEGGER_RTT_printf(0, "Balancing complete. \n");
    iCount = 0;
  }
}

void startCellBalance()
{
  SEGGER_RTT_printf(0, " #### BMS Cell Balance #### \n");
  
  for(int i = 0; i < TOTAL_CELL - 1; i++)
  {
     if(i == S6) continue;
     
     if(bStatusFET[i])  dischargerSet(i + 1);
     else               SEGGER_RTT_printf(0, "Cell[%d] OK. \n", i + 1);
  }
}


int8_t iCountSPI = 0;
void measurement_loop(uint8_t datalog_en)
{
//  int8_t error = 0;  // not use
  if (WRITE_CONFIG == ENABLED)
  {
    wakeup_sleep(TOTAL_IC);
    LTC6811_wrcfg(TOTAL_IC,bms_ic);
  }

  if (READ_CONFIG == ENABLED)
  {
    wakeup_sleep(TOTAL_IC);
//    error = LTC6811_rdcfg(TOTAL_IC,bms_ic);
    LTC6811_rdcfg(TOTAL_IC,bms_ic);
//    check_error(error);
  }

  
  if (MEASURE_CELL == ENABLED)
  {
    
      wakeup_idle(TOTAL_IC);
      LTC6811_adcv(ADC_CONVERSION_MODE,ADC_DCP,CELL_CH_TO_CONVERT);

      for (volatile int delay=0; delay<0x0E5; delay++);
      LTC6811_pollAdc();
      for (volatile int delay=0; delay<0x2200; delay++);
      wakeup_idle(TOTAL_IC);

      ////    error = LTC6811_rdcv(NO_OF_REG, TOTAL_IC,bms_ic);
      LTC6811_rdcv(NO_OF_REG, TOTAL_IC,bms_ic);
      ////    check_error(error);
  }
  
  
  if (MEASURE_AUX == ENABLED)
  {
    wakeup_idle(TOTAL_IC);
    LTC6811_adax(ADC_CONVERSION_MODE , AUX_CH_ALL);
    for (volatile int delay=0; delay<0x070; delay++);
    LTC6811_pollAdc();
    for (volatile int delay=0; delay<0x1000; delay++);
    wakeup_idle(TOTAL_IC);
//    error = LTC6811_rdaux(NO_OF_REG,TOTAL_IC,bms_ic); // Set to read back all aux registers
    LTC6811_rdaux(NO_OF_REG,TOTAL_IC,bms_ic); // Set to read back all aux registers
//    check_error(error);
    
//    readTemp();
  }

  if (MEASURE_STAT == ENABLED)
  {
    wakeup_idle(TOTAL_IC);
    LTC6811_adstat(ADC_CONVERSION_MODE, STAT_CH_ALL);
    LTC6811_pollAdc();
    wakeup_idle(TOTAL_IC);
//    error = LTC6811_rdstat(NO_OF_REG,TOTAL_IC,bms_ic); // Set to read back all aux registers
     LTC6811_rdstat(NO_OF_REG,TOTAL_IC,bms_ic); // Set to read back all aux registers
//    check_error(error);
  }

  if (PRINT_PEC == ENABLED)
  {
//    print_pec();
  }
}


void readStatusBMS()
{
    char chRead ;
    uint8_t strLength = 1, j;
    
    if((chRead = SEGGER_RTT_GetKey()) != 0xFF)
    {
      char arrCommand[5] = {chRead};
      
      while(chRead != '\n')
      {
        chRead = SEGGER_RTT_GetKey(); 
        arrCommand[strLength] = chRead;
        strLength++;
      }
      
      
      if(strLength == 5)        sprintf(arrCommand, "%c%c%c", arrCommand[0], arrCommand[1], arrCommand[2]);     // add char
      else if(strLength == 4)   sprintf(arrCommand, "%c%c", arrCommand[0], arrCommand[1]);                      // add char
      else if(strLength <= 3)   sprintf(arrCommand, "%c", arrCommand[0]);                                       // add char
      else {}
      

      char str[strLength - 2]; 
      for(j = 0; j < strLength - 2; j++)
      {
        str[j] = arrCommand[j];
      }
      
      if(strncmp(str, "v", j) == 0)       readCell();          // strncmp = if equal return 0   
//      if(strncmp(str, "t", j) == 0)       readTemp();          
      else if(strncmp(str, "d1", j) == 0) dischargerSet(1);
      else if(strncmp(str, "d2", j) == 0) dischargerSet(2);
      else if(strncmp(str, "d3", j) == 0) dischargerSet(3);
      else if(strncmp(str, "d4", j) == 0) dischargerSet(4);
      else if(strncmp(str, "d5", j) == 0) dischargerSet(5);
      else if(strncmp(str, "d6", j) == 0) dischargerSet(7);
      else if(strncmp(str, "d7", j) == 0) dischargerSet(8);
      else if(strncmp(str, "d8", j) == 0) dischargerSet(9);
      else if(strncmp(str, "d9", j) == 0) dischargerSet(10);
      else if(strncmp(str, "d10", j) == 0) dischargerSet(11);
      else if(strncmp(str, "da", j) == 0) 
      {
        dischargerSet(1);
        dischargerSet(3);
        dischargerSet(5);
      }
      else if(strncmp(str, "db", j) == 0) 
      {
        dischargerSet(7);
        dischargerSet(8);
        dischargerSet(10);
      }
      
      else if(strncmp(str, "b", j) == 0) 
      {
        bBalanceStatus = !bBalanceStatus;
        iCount = 0;
        if(!bBalanceStatus) SEGGER_RTT_printf(0, " #### BMS Cell Balance Off #### \n");
        else                SEGGER_RTT_printf(0, " #### BMS Cell Balance On #### \n");
      }
      else if(strncmp(str, "c", j) == 0) dischargerClear();
      else if(strncmp(str, "h", j) == 0) commandInfo();
      else { }
      

      strLength = 0;
    }
    
}


void readCell()
{  
    SEGGER_RTT_printf(0, " #### BMS Cell data #### \n");
    for(int i = 0; i < TOTAL_CELL; i++)
    {
      static char chArr[10];
      sprintf(chArr, "Cell[%d] : %.4f [V]",i + 1,  bms_ic[0].cells.c_codes[i] * (float)0.0001);
      SEGGER_RTT_printf(0, "%s \n", chArr);
    }
}


void commandInfo()
{
  
  SEGGER_RTT_printf(0, "#### Command Information. ####\n");
  SEGGER_RTT_printf(0, "'h' : help\n'd1'~'d10' : Discharge FET\n'c' : Reset FET\n'v' : Cell Voltage\n'b' : Start balancing");
}


//
//
//void readTemp()
//{
//    
//////    SEGGER_RTT_printf(0, "------- BMS Temp data ------- \n");
////    static uint16_t Vo[6];
////    double R1 = 203750;
////    double logR2, R2;
////    static double T[6];
////    double c1 = 0.7715923754e-3, c2 = 2.707271732-4, c3 = 1.138266460e-7;
////
////    
////    for(int i = 0; i < 6; i++)
////    {  // 0, 1, 3, 4 Temp, 2 current, 5 Vref
////       
////       if(i == 2 || i == 5)
////       {
////         continue;
////       }
////       
////       Vo[i] = bms_ic[0].aux.a_codes[i];
////       R2 = R1 * (65534 / ((double)Vo[i]) - 1.0);
////       logR2 = log(R2);
////       
////       T[i] = (1.0 / (c1 + (c2 * logR2) + (c3 * logR2 * logR2 * logR2)));
////       T[i] = T[i] - 273.15;
//////       T[i] = (T[i] * 9.0)/5.0 + 32;
////       
////    }
//}


void LTC6811_test()
{
   measurement_loop(0);
   
   
   // Cell balance
    if(bBalanceStatus)
    {
      nrf_delay_ms(10);
      if(iCount == 0 )      dischargerClear();   
      else if(iCount == 2)  checkCell();
      else if(iCount == 3)  dischargerClear(); 
      else if(iCount == 5)  startCellBalance(); 
      else if(iCount == 1000)  
      {
        iCount = -1;
      }
//      SEGGER_RTT_printf(0, "%d\n", iCount);
      iCount++;
    }
    // -----------------
}




//static uint16_t uiTempADC[] = {
//3066, 3243, 3428, 3621, 3821, 4029, 4245, 4470, 4703, 4945, 5197, 5457, 5727, 6007, 6297, 6596, 6906, 
//7226, 7557, 7898, 8250, 8613, 8986, 9371, 9766, 10172, 10589, 11017, 11456, 11905, 12365, 12835, 13316, 
//13943, 14307, 14818, 15338, 15867, 16404, 16951, 17470, 18068, 18638, 19214, 19798, 20388, 20984, 21584, 
//22190, 22801, 23392, 24032, 24653, 25275, 25901, 26528, 27154, 27783, 28410, 29038, 29664, 30289, 30914, 
//31534, 32153, 32767, 33380, 33987, 34590, 35190, 35784, 36371, 36954, 37532, 38103, 38668, 39225, 39775, 
//40319, 40857, 41386, 41907, 42420, 42925, 43426, 43983, 44397, 44871, 45337, 45793, 46255, 46683, 47116, 
//47540, 47958, 48365, 48764, 49155, 49538, 49915, 50283, 50641, 50995, 51339, 51675, 52003, 52327, 52642, 
//52948, 53249, 53541, 53827, 54107, 54381, 54648, 54909, 55163, 55411, 55650, 55888, 56117, 56339, 56558, 
//56774, 56981, 57180, 57380, 57572, 57760, 57943, 58123, 58294, 58466, 58628, 58791, 58949, 59098, 59248, 
//59393, 59533, 59669, 59805, 59936, 60062, 60184, 60306
//};
//
//
//static int16_t uiTempVal[] = {
//-40, -39, -38, -37, -36, -35, -34, -33, -32, -31, -30, -29, -28, -27, -26, -25, -24, -23, -22, -21, -20, -19, -18, -17, -16, 
//-15, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 
//11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 
//31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44,45, 46, 47, 48, 49, 50, 
//51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 
//71, 72, 73, 74, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 
//91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105 
//};